from ..credentials import PlatformCredentials


class LoginCommand:
    def __init__(self, args):
        self.args = args

    def run(self):
        credentials = PlatformCredentials()
        credentials.store(self.args)



"""Fixes for Maven projects"""
import os
import os.path
import platform
import subprocess
import tempfile
import xml.etree.ElementTree as ET

import requests

import instarepo.fixers.context
import instarepo.git
import instarepo.github
import instarepo.xml_utils
from instarepo.fixers.base import MissingFileFix
from .finders import is_maven_project
from .readme import locate_badges, merge_badges


MAVEN_YML = """# This workflow will build a Java project with Maven, and cache/restore any dependencies to improve the workflow execution time
# For more information see: https://help.github.com/actions/language-and-framework-guides/building-and-testing-java-with-maven

name: Java CI with Maven

on:
  push:
    branches: [ trunk ]
  pull_request:
    branches: [ trunk ]

jobs:
  build:

    runs-on: ubuntu-latest

    steps:
    - uses: actions/checkout@v2
    - name: Set up JDK 11
      uses: actions/setup-java@v2
      with:
        java-version: '11'
        distribution: 'adopt'
        cache: maven
    - name: Build with Maven
      run: mvn -B package --file pom.xml
"""


class MustHaveCIFix(MissingFileFix):
    """If missing, adds a GitHub action Maven build workflow"""

    def get_filename(self):
        return ".github/workflows/maven.yml"

    def should_process_repo(self):
        return is_maven_project(self.context.git.dir)

    def get_contents(self):
        return MAVEN_YML.replace("trunk", self.context.default_branch())


class MavenBadgesFix:
    """
    Fixes badges for Maven libraries.

    Does not work for local git repositories.
    """

    def __init__(self, context: instarepo.fixers.context.Context):
        self.context = context
        self.maven = Maven(context.git.dir)

    def run(self):
        if not self.context.git.isfile("README.md"):
            return []
        if not self.context.git.isfile("pom.xml"):
            return []
        if not self.context.repo:
            return []
        badges = self._badges_dict(self.context.repo)

        if not badges:
            return []

        with open(self.context.git.join("README.md"), "r", encoding="utf-8") as file:
            before_badges, existing_badges, after_badges = locate_badges(file.read())
            has_changes = False
            for i in range(len(existing_badges)):
                existing_badge = existing_badges[i]
                for needle, markdown in list(badges.items()):
                    if needle in existing_badge:
                        # delete the badge from the dictionary
                        # so that we don't add it again later
                        del badges[needle]
                        if existing_badge != markdown:
                            has_changes = True
                            existing_badges[i] = markdown
                        break
            if badges:
                existing_badges.extend(badges.values())
                has_changes = True
        if not has_changes:
            return []
        with open(self.context.git.join("README.md"), "w", encoding="utf-8") as file:
            file.write(merge_badges(before_badges, existing_badges, after_badges))
        self.context.git.add("README.md")
        msg = "Updated Maven badges in README.md"
        self.context.git.commit(msg)
        return [msg]

    def _badges_dict(self, repo: instarepo.github.Repo):
        badges = {}
        badges.update(self._github_actions_badge(repo))
        badges.update(self._effective_pom_badges())
        return badges

    def _github_actions_badge(self, repo: instarepo.github.Repo):
        if self.context.git.isfile(".github", "workflows", "maven.yml"):
            needle = "actions/workflows"
            markdown = f"[![Java CI with Maven](https://github.com/{repo.full_name}/actions/workflows/maven.yml/badge.svg)](https://github.com/{repo.full_name}/actions/workflows/maven.yml)"
            return {needle: markdown}
        return {}

    def _effective_pom_badges(self):
        badges = {}
        with tempfile.TemporaryDirectory() as tmp_dir:
            filename = os.path.join(tmp_dir, "pom.xml")
            self.maven.run("-B", "help:effective-pom", f"-Doutput={filename}")
            tree = instarepo.xml_utils.parse(filename)
            root = tree.getroot()
            badges.update(maven_central_badge(root))
            # edge case for checkstyle-rules artifact which has to publish
            # javadoc but doesn't have any source code
            if self.context.git.isdir("src", "main", "java"):
                badges.update(javadoc_badge(root))
        return badges


def maven_central_badge(root: ET.Element):
    group_id = root.findtext("{http://maven.apache.org/POM/4.0.0}groupId")
    if not group_id:
        return {}
    artifact_id = root.findtext("{http://maven.apache.org/POM/4.0.0}artifactId")
    if not artifact_id:
        return {}
    group_id_as_path = group_id.replace(".", "/")
    url = f"https://repo1.maven.org/maven2/{group_id_as_path}/{artifact_id}/"
    response = requests.get(url)
    if not response.ok:
        return {}
    needle = "maven-central"
    markdown = f"[![Maven Central](https://img.shields.io/maven-central/v/{group_id}/{artifact_id}.svg?label=Maven%20Central)](https://search.maven.org/search?q=g:%22{group_id}%22%20AND%20a:%22{artifact_id}%22)"
    return {needle: markdown}


def javadoc_badge(root: ET.Element):
    plugins = instarepo.xml_utils.find(
        root,
        "{http://maven.apache.org/POM/4.0.0}build",
        "{http://maven.apache.org/POM/4.0.0}plugins",
    )
    if plugins is None:
        return {}
    uses_maven_source_plugin = False
    uses_maven_javadoc_plugin = False
    for plugin in plugins:
        plugin_artifact_id = plugin.findtext(
            "{http://maven.apache.org/POM/4.0.0}artifactId"
        )
        if plugin_artifact_id == "maven-source-plugin":
            if "jar-no-fork" in plugin_goals(plugin):
                uses_maven_source_plugin = True
        if plugin_artifact_id == "maven-javadoc-plugin":
            if "jar" in plugin_goals(plugin):
                uses_maven_javadoc_plugin = True
    if not (uses_maven_source_plugin and uses_maven_javadoc_plugin):
        return {}
    group_id = root.findtext("{http://maven.apache.org/POM/4.0.0}groupId")
    if not group_id:
        return {}
    artifact_id = root.findtext("{http://maven.apache.org/POM/4.0.0}artifactId")
    if not artifact_id:
        return {}
    needle = "javadoc"
    markdown = f"[![javadoc](https://javadoc.io/badge2/{group_id}/{artifact_id}/javadoc.svg)](https://javadoc.io/doc/{group_id}/{artifact_id})"
    return {needle: markdown}


def plugin_goals(plugin: ET.Element):
    executions = plugin.find("{http://maven.apache.org/POM/4.0.0}executions")
    if executions is None:
        return set()
    executions = executions.findall("{http://maven.apache.org/POM/4.0.0}execution")
    goals = map(
        lambda x: x.find("{http://maven.apache.org/POM/4.0.0}goals"),
        executions,
    )
    goals = [x for x in goals if x]
    return set(
        y.text
        for x in goals
        for y in x.findall("{http://maven.apache.org/POM/4.0.0}goal")
        if y.text
    )


class Maven:
    """
    Runs Maven commands.
    """

    def __init__(self, directory: str):
        """
        Creates an instance of this class.

        :param directory: The directory of a Maven project
        """
        self.directory = directory

    # pylint: disable=subprocess-run-check
    def run(self, *args) -> str:
        """
        Runs Maven commands.
        """
        mvn_executable = "mvn.cmd" if platform.system() == "Windows" else "mvn"
        result = subprocess.run(
            [mvn_executable, *args],
            cwd=self.directory,
            encoding="utf-8",
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        maven_output = result.stdout.strip()
        if result.returncode != 0:
            raise ChildProcessError(maven_output)
        return maven_output

    def sort_pom(self):
        self.run(
            "-B",
            "com.github.ekryd.sortpom:sortpom-maven-plugin:sort",
            "-Dsort.createBackupFile=false",
        )


class UrlFix:
    """
    Ensures Maven projects have the correct URL and SCM sections.

    Does not work for local git repositories.
    """

    def __init__(self, context: instarepo.fixers.context.Context):
        self.context = context

    def run(self):
        if not self.context.git.isfile("pom.xml") or not self.context.repo:
            return []
        try:
            ET.register_namespace("", "http://maven.apache.org/POM/4.0.0")
            return self.do_run(self.context.repo)
        finally:
            ET.register_namespace("maven", "http://maven.apache.org/POM/4.0.0")

    def do_run(self, repo: instarepo.github.Repo):
        has_changes = False
        tree = instarepo.xml_utils.parse(self.context.git.join("pom.xml"))
        root = tree.getroot()
        has_changes |= ensure_element(
            root, "{http://maven.apache.org/POM/4.0.0}url", repo.html_url
        )
        scm = root.find("{http://maven.apache.org/POM/4.0.0}scm")
        if scm is None:
            has_changes = True
            scm = ET.Element("{http://maven.apache.org/POM/4.0.0}scm")
            root.append(scm)
        has_changes |= ensure_element(
            scm,
            "{http://maven.apache.org/POM/4.0.0}connection",
            f"scm:git:{repo.clone_url}",
        )
        has_changes |= ensure_element(
            scm,
            "{http://maven.apache.org/POM/4.0.0}developerConnection",
            f"scm:git:{repo.ssh_url}",
        )
        has_changes |= ensure_element(
            scm, "{http://maven.apache.org/POM/4.0.0}tag", "HEAD"
        )
        has_changes |= ensure_element(
            scm, "{http://maven.apache.org/POM/4.0.0}url", repo.html_url
        )
        if has_changes:
            tree.write(self.context.git.join("pom.xml"))
            Maven(self.context.git.dir).sort_pom()
            self.context.git.add("pom.xml")
            msg = "Corrected url info in pom.xml"
            self.context.git.commit(msg)
            return [msg]

        return []


def ensure_element(parent: ET.Element, child_name: str, child_text: str) -> bool:
    has_changes = False
    child_element = parent.find(child_name)
    if child_element is None:
        has_changes = True
        child_element = ET.Element(child_name)
        parent.append(child_element)
    if child_element.text != child_text:
        has_changes = True
        child_element.text = child_text
    return has_changes



"""Unit tests for maven.py"""
import xml.etree.ElementTree as ET

import instarepo.xml_utils
from .maven import javadoc_badge


def test_detect_javadoc_usage_in_pom():
    contents = """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 https://maven.apache.org/xsd/maven-4.0.0.xsd">
  <modelVersion>4.0.0</modelVersion>
  <groupId>com.github.ngeor</groupId>
  <artifactId>yak4j-spring-test-utils</artifactId>
  <version>0.22.0-SNAPSHOT</version>
  <name>yak4j-spring-test-utils</name>
  <build>
    <plugins>
      <plugin>
        <artifactId>maven-checkstyle-plugin</artifactId>
        <version>3.1.2</version>
        <executions>
          <execution>
            <id>validate</id>
            <phase>validate</phase>
            <goals>
              <goal>check</goal>
            </goals>
          </execution>
        </executions>
      </plugin>
      <plugin>
        <artifactId>maven-source-plugin</artifactId>
        <version>3.2.1</version>
        <executions>
          <execution>
            <id>attach-sources</id>
            <goals>
              <goal>jar-no-fork</goal>
            </goals>
          </execution>
        </executions>
      </plugin>
      <plugin>
        <artifactId>maven-javadoc-plugin</artifactId>
        <version>3.1.1</version>
        <executions>
          <execution>
            <id>attach-javadocs</id>
            <goals>
              <goal>jar</goal>
            </goals>
          </execution>
        </executions>
      </plugin>
      <plugin>
        <artifactId>maven-surefire-plugin</artifactId>
        <version>3.0.0-M5</version>
        <executions>
          <execution>
            <id>default-test</id>
            <phase>test</phase>
            <goals>
              <goal>test</goal>
            </goals>
          </execution>
        </executions>
      </plugin>
    </plugins>
  </build>
</project>
    """
    root = ET.fromstring(contents, parser=instarepo.xml_utils.create_parser())
    assert root is not None
    badges = javadoc_badge(root)
    assert badges == {
        "javadoc": "[![javadoc](https://javadoc.io/badge2/com.github.ngeor/yak4j-spring-test-utils/javadoc.svg)](https://javadoc.io/doc/com.github.ngeor/yak4j-spring-test-utils)"
    }



"""Unit tests for dotnet.py"""

import xml.etree.ElementTree as ET
from instarepo.xml_utils import create_parser
from .dotnet import (
    get_projects_from_sln_file_contents,
    comment,
    SlnParser,
    csproj_root_node_to_os,
)


def test_get_projects_from_sln_file_contents():
    """Tests extracting csproj files from sln contents"""
    sln_file = """# Visual Studio 15
VisualStudioVersion = 15.0.27130.2020
MinimumVisualStudioVersion = 10.0.40219.1
Project("{FAE04EC0-301F-11D3-BF4B-00C04F79EFBC}") = "CVRender.Web", "CVRender.Web\\CVRender.Web.csproj", "{46D05687-EB9B-4885-9A14-1BDC8BBB253B}"
EndProject
Project("{FAE04EC0-301F-11D3-BF4B-00C04F79EFBC}") = "CVRender", "CVRender\\CVRender.csproj", "{BD17C766-DF9E-4117-A8CB-2BAA8FE6D9B9}"
EndProject
Project("{FAE04EC0-301F-11D3-BF4B-00C04F79EFBC}") = "CVRender.Tests", "CVRender.Tests\\CVRender.Tests.csproj", "{FDABDD4B-8BF5-4E4A-B977-400D0CE04D4A}"
EndProject
Project("{2150E333-8FDC-42A3-9474-1A3956D46DE8}") = "Solution Items", "Solution Items", "{184005A4-82D4-489E-BD9C-390DFEBC074D}"
	ProjectSection(SolutionItems) = preProject
		.gitignore = .gitignore
		appveyor.yml = appveyor.yml
		LICENSE = LICENSE
		README.md = README.md
	EndProjectSection
EndProject"""
    projects = list(get_projects_from_sln_file_contents(sln_file))
    assert projects == [
        "CVRender.Web\\CVRender.Web.csproj",
        "CVRender\\CVRender.csproj",
        "CVRender.Tests\\CVRender.Tests.csproj",
    ]


def test_get_projects_from_sln_file_contents_new_guid():
    """Tests support for the new GUID used by the dotnet CLI"""
    sln_file = """
Microsoft Visual Studio Solution File, Format Version 12.00
# Visual Studio Version 16
VisualStudioVersion = 16.0.29102.190
MinimumVisualStudioVersion = 15.0.26124.0
Project("{9A19103F-16F7-4668-BE54-9A1E7A4F7556}") = "DuplicateFileFinder", "DuplicateFileFinder\\DuplicateFileFinder.csproj", "{55E0401D-7FA2-4809-B98F-3D5C1EEEF336}"
EndProject
Project("{9A19103F-16F7-4668-BE54-9A1E7A4F7556}") = "PostfixNumberRemover", "PostfixNumberRemover\\PostfixNumberRemover.csproj", "{4884CCE8-C949-41FD-8136-BCCB2AF657E6}"
EndProject
Project("{9A19103F-16F7-4668-BE54-9A1E7A4F7556}") = "PostfixNumberRemover.Tests", "PostfixNumberRemover.Tests\\PostfixNumberRemover.Tests.csproj", "{F871B937-163D-400D-9BAC-D49C8BC7A629}"
EndProject"""
    projects = list(get_projects_from_sln_file_contents(sln_file))
    assert projects == [
        "DuplicateFileFinder\\DuplicateFileFinder.csproj",
        "PostfixNumberRemover\\PostfixNumberRemover.csproj",
        "PostfixNumberRemover.Tests\\PostfixNumberRemover.Tests.csproj",
    ]


def test_comment():
    parsed, remaining = comment()("# hello")
    assert parsed == "# hello"
    assert remaining == ""
    parsed, remaining = comment()("# hello\r\n\r\ngood bye")
    assert parsed == "# hello"
    assert remaining == "good bye"


def test_sln_parser():
    sln_file = """# Visual Studio 15
VisualStudioVersion = 15.0.27130.2020
MinimumVisualStudioVersion = 10.0.40219.1
Project("{FAE04EC0-301F-11D3-BF4B-00C04F79EFBC}") = "CVRender.Web", "CVRender.Web\\CVRender.Web.csproj", "{46D05687-EB9B-4885-9A14-1BDC8BBB253B}"
EndProject"""
    parser = SlnParser(sln_file)
    assert parser.next() == "# Visual Studio 15"
    assert parser.next() == "VisualStudioVersion"
    assert parser.next() == "="
    assert parser.next() == "15.0.27130.2020"
    assert parser.next() == "\n"
    assert parser.next() == "MinimumVisualStudioVersion"
    assert parser.next() == "="
    assert parser.next() == "10.0.40219.1"
    assert parser.next() == "\n"
    assert parser.next() == "Project"
    assert parser.next() == "("
    assert parser.next() == '"{FAE04EC0-301F-11D3-BF4B-00C04F79EFBC}"'
    assert parser.next() == ")"
    assert parser.next() == "="
    assert parser.next() == '"CVRender.Web"'
    assert parser.next() == ","
    assert parser.next() == '"CVRender.Web\\CVRender.Web.csproj"'
    assert parser.next() == ","
    assert parser.next() == '"{46D05687-EB9B-4885-9A14-1BDC8BBB253B}"'
    assert parser.next() == "\n"
    assert parser.next() == "EndProject"


def test_parse_msbuild_style_csproj():
    contents = """<?xml version="1.0" encoding="utf-8"?>
<Project ToolsVersion="15.0" DefaultTargets="Build" xmlns="http://schemas.microsoft.com/developer/msbuild/2003">
</Project>
"""
    root = ET.fromstring(contents, create_parser())
    assert csproj_root_node_to_os(root) == "windows"


def test_parse_dotnet_style_csproj():
    contents = """<Project Sdk="Microsoft.NET.Sdk">

  <PropertyGroup>
    <TargetFramework>netstandard2.1</TargetFramework>
  </PropertyGroup>
</Project>
"""
    root = ET.fromstring(contents, create_parser())
    assert csproj_root_node_to_os(root) == "linux"


def test_parse_dotnet_net47_style_csproj():
    contents = """<Project Sdk="Microsoft.NET.Sdk">

  <PropertyGroup>
    <TargetFramework>net47</TargetFramework>
  </PropertyGroup>
</Project>
"""
    root = ET.fromstring(contents, create_parser())
    assert csproj_root_node_to_os(root) == "windows"



"""
Applies fixes to a repository that is either locally checked out
or remote on GitHub.
"""
import logging
import tempfile
from typing import Iterable

import instarepo.git
import instarepo.github
import instarepo.repo_source

from ..credentials import build_requests_auth
from ..fixers.discovery import (
    all_fixer_classes,
    select_fixer_classes,
)
from ..fixers.naming import fixer_class_to_fixer_key


class FixCommand:
    """
    Applies fixes to a repository that is either locally checked out
    or remote on GitHub.
    """

    def __init__(self, args):
        if args.local_dir:
            self.delegate = FixLocal(args)
        else:
            self.delegate = FixRemote(args)

    def run(self):
        self.delegate.run()


class FixBase:
    """
    Base class for common features among FixLocal and FixRemote.
    """

    def __init__(self, args):
        self.dry_run: bool = args.dry_run
        self.verbose: bool = args.verbose
        self.fixer_classes = select_fixer_classes(args.only_fixers, args.except_fixers)
        self.config = instarepo.fixers.config.load_config(args.config_file)

    def run(self):
        if not self.fixer_classes:
            logging.error("No fixers selected!")
            return
        logging.debug(
            "Using fixers %s",
            ", ".join(map(fixer_class_to_fixer_key, self.fixer_classes)),
        )


class FixLocal(FixBase):
    """
    Applies fixes to a locally checked-out repository.
    """

    def __init__(self, args):
        super().__init__(args)
        if not args.local_dir:
            raise ValueError("local_dir must be specified")
        self.local_dir = args.local_dir

    def run(self):
        super().run()
        logging.info("Processing local repo %s", self.local_dir)
        git = instarepo.git.GitWorkingDir(self.local_dir, quiet=not self.verbose)
        context = instarepo.fixers.context.Context(
            git=git, config=self.config, verbose=self.verbose
        )
        if context.get_setting("enabled"):
            composite_fixer = create_composite_fixer(self.fixer_classes, context)
            composite_fixer.run()


BRANCH_NAME = "instarepo_branch"


class FixRemote(FixBase):
    """
    Applies fixes to a GitHub repository.
    """

    def __init__(self, args):
        super().__init__(args)
        if args.local_dir:
            raise ValueError("local_dir must be empty")
        auth = build_requests_auth(args)
        if args.dry_run:
            self.github = instarepo.github.GitHub(auth=auth)
        else:
            self.github = instarepo.github.ReadWriteGitHub(auth=auth)
        self.auto_merge = args.auto_merge
        self.force = args.force
        self.repo_source = (
            instarepo.repo_source.RepoSourceBuilder()
            .with_github(self.github)
            .with_args(args)
            .build()
        )

    def run(self):
        super().run()
        repos = self.repo_source.get()
        for repo in repos:
            if self.config.get_setting(repo.full_name, "enabled"):
                self._process(repo)

    def _process(self, repo: instarepo.github.Repo):
        logging.info("Processing repo %s", repo.name)
        with tempfile.TemporaryDirectory() as tmpdirname:
            logging.debug("Cloning repo into temp dir %s", tmpdirname)
            git = instarepo.git.clone(repo.ssh_url, tmpdirname, quiet=not self.verbose)
            self._process_in_temp_directory(repo, git)

    def _process_in_temp_directory(
        self, repo: instarepo.github.Repo, git: instarepo.git.GitWorkingDir
    ):
        is_remote_branch_present = git.is_remote_branch_present(BRANCH_NAME)
        needs_force_push = False
        behind = 0
        ahead = 0
        if is_remote_branch_present:
            behind, ahead = git.get_behind_ahead(
                f"origin/{repo.default_branch}", f"origin/{BRANCH_NAME}"
            )
            logging.debug(
                "Remote branch exists and is %d commits behind and %d commits ahead of %s",
                behind,
                ahead,
                repo.default_branch,
            )
            if self._should_start_from_scratch(behind, ahead, git):
                git.create_branch(BRANCH_NAME)
                needs_force_push = True
                behind = 0
                ahead = 0
            else:
                git.checkout(BRANCH_NAME)
        else:
            git.create_branch(BRANCH_NAME)
        context = instarepo.fixers.context.Context(
            git=git,
            config=self.config,
            repo=repo,
            github=self.github,
            verbose=self.verbose,
        )
        composite_fixer = create_composite_fixer(
            self.fixer_classes,
            context,
        )
        changes = composite_fixer.run()
        if changes:
            self._create_merge_request(repo, git, changes, needs_force_push)
        elif ahead > 0 and self.auto_merge:
            # no changes in this run, but we are ahead of default branch, we can auto-merge
            merged = self._auto_merge_existing_mr(repo)
            if is_remote_branch_present and merged:
                if self.dry_run:
                    logging.info("Would have deleted remote branch")
                else:
                    git.delete_remote_branch(BRANCH_NAME)
        elif ahead == 0:
            # no changes and at the same point as the default branch, we can auto-close the MR
            if is_remote_branch_present:
                if self.dry_run:
                    logging.info("Would have deleted remote branch")
                else:
                    git.delete_remote_branch(BRANCH_NAME)
            self._close_mr_if_exists(repo)

    def _create_merge_request(
        self,
        repo: instarepo.github.Repo,
        git: instarepo.git.GitWorkingDir,
        changes: Iterable[str],
        needs_force_push: bool,
    ):
        if self.dry_run:
            logging.info("Would have created PR for repo %s", repo.name)
            return
        git.push(force=needs_force_push)
        if len(self._list_merge_requests(repo)) > 0:
            logging.info("PR already exists for repo %s", repo.name)
        else:
            html_url = self.github.create_merge_request(
                repo.full_name,
                BRANCH_NAME,
                repo.default_branch,
                "instarepo automatic PR",
                format_body(changes),
            )
            logging.info("Created PR for repo %s - %s", repo.name, html_url)

    def _list_merge_requests(self, repo: instarepo.github.Repo):
        head = self.github.auth.username + ":" + BRANCH_NAME
        return self.github.list_merge_requests(
            repo.full_name, head, repo.default_branch
        )

    def _close_mr_if_exists(self, repo: instarepo.github.Repo):
        merge_requests = self._list_merge_requests(repo)
        if not merge_requests:
            return
        merge_request = merge_requests[0]
        number = merge_request["number"]
        self.github.create_issue_comment(
            repo.full_name,
            number,
            "It seems the changes in this MR have already been fixed, auto-closing.",
        )
        self.github.close_merge_request(repo.full_name, number)

    def _auto_merge_existing_mr(self, repo: instarepo.github.Repo):
        merge_requests = self._list_merge_requests(repo)
        for merge_request in merge_requests:
            if self._auto_merge_one_existing_mr(repo, merge_request):
                return True
        return False

    def _auto_merge_one_existing_mr(self, repo: instarepo.github.Repo, merge_request):
        number = merge_request["number"]
        details = self.github.get_merge_request(repo.full_name, number)
        if not details["mergeable"]:
            logging.debug("Cannot merge MR because GitHub reports it is not mergeable")
            return False
        mergeable_state = details["mergeable_state"]
        if mergeable_state != "clean":
            logging.debug(
                "Cannot merge MR because the mergeable state is not clean but %s",
                mergeable_state,
            )
            return False
        head_sha = merge_request["head"]["sha"]
        check_runs = self.github.list_check_runs(repo.full_name, head_sha)
        if check_runs["total_count"] <= 0:
            logging.debug("Cannot merge MR because there are no check runs")
            return False
        for check_run in check_runs["check_runs"]:
            status = check_run["status"]
            if status != "completed":
                logging.debug("Cannot merge MR because there are incomplete check runs")
                return False
            conclusion = check_run["conclusion"]
            if conclusion != "success":
                logging.debug(
                    "Cannot merge MR because there are unsuccessful check runs"
                )
                return False

        self.github.merge_merge_request(repo.full_name, number)
        return True

    def _should_start_from_scratch(
        self, behind, ahead, git: instarepo.git.GitWorkingDir
    ):
        if self.force:
            logging.info("Force flag is set, starting branch from scratch")
            return True
        if behind > 0:
            logging.info(
                "Remote branch is behind default branch, starting from scratch"
            )
            return True
        if ahead > 0:
            author_names = git.get_author_names(f"origin/{BRANCH_NAME}")
            if len(author_names) != ahead:
                logging.warning(
                    "Found %d author names but expected %d", len(author_names), ahead
                )
                return False
            has_custom_authors = any(
                filter(lambda x: x != instarepo.git.AUTHOR_NAME, author_names)
            )
            if has_custom_authors:
                logging.info(
                    "Found commits by different authors, keeping existing branch"
                )
                return False
            logging.info("All commits are from instarepo, starting from scratch")
            return True
        logging.info("Existing branch has no commits")
        return False


def create_composite_fixer(fixer_classes, context: instarepo.fixers.context.Context):
    return instarepo.fixers.base.CompositeFix(
        list(
            map(
                lambda fixer_class: fixer_class(context),
                fixer_classes,
            )
        )
    )


def format_body(changes: Iterable[str]) -> str:
    body = "The following fixes have been applied:\n"
    for change in changes:
        lines = _non_empty_lines(change)
        first = True
        for line in lines:
            if first:
                body += "- "
                first = False
            else:
                body += "  "
            body += line + "\n"
    return body


def _non_empty_lines(value: str) -> Iterable[str]:
    lines = value.split("\n")
    stripped_lines = (line.strip() for line in lines)
    return (line for line in stripped_lines if line)


def epilog():
    """
    Creates a help text for the available fixers.
    """
    result = ""
    for clz in all_fixer_classes():
        result += fixer_class_to_fixer_key(clz)
        result += "\n    "
        result += clz.__doc__
        result += "\n"
    return result



import functools
from typing import Optional
import instarepo.git
import instarepo.github
import instarepo.fixers.config


class Context:
    def __init__(
        self,
        git: instarepo.git.GitWorkingDir,
        config: instarepo.fixers.config.Config,
        repo: Optional[instarepo.github.Repo] = None,
        github: Optional[instarepo.github.GitHub] = None,
        verbose: bool = False,
    ):
        self.git = git
        self.config = config
        self.repo = repo
        self.github = github
        self.verbose = verbose

    @functools.lru_cache()
    def full_name(self):
        """
        Gets the full name of the repository.
        If the GitHub metadata is available in the 'repo' field, that is returned directly.
        Otherwise, i.e. for locally checked out repos, the method tries to extract the
        full name from the git remote URL.
        """
        if self.repo:
            return self.repo.full_name
        remote_url = self.git.get_remote_url()
        prefix = "git@github.com:"
        if remote_url.startswith(prefix):
            result = remote_url[len(prefix) :]
        else:
            raise ValueError(
                f"Unsupported git remote {remote_url} did not start with {prefix}"
            )
        suffix = ".git"
        if result.endswith(suffix):
            result = result[: -len(suffix)]
        else:
            raise ValueError(
                f"Unsupported git remote {remote_url} did not end with {suffix}"
            )
        return result

    def get_setting(self, *args):
        return self.config.get_setting(self.full_name(), *args)

    @functools.lru_cache()
    def default_branch(self):
        if self.repo:
            return self.repo.default_branch
        return self.git.get_default_branch()



from typing import List

import instarepo.fixers.base
import instarepo.fixers.changelog
import instarepo.fixers.ci
import instarepo.fixers.config
import instarepo.fixers.context
import instarepo.fixers.dotnet
import instarepo.fixers.license
import instarepo.fixers.maven
import instarepo.fixers.missing_files
import instarepo.fixers.pascal
import instarepo.fixers.readme
import instarepo.fixers.repo_description
import instarepo.fixers.vb6

from .naming import FIXER_SUFFIX, fixer_class_to_fixer_key


def try_get_fixer_order(fixer_class):
    return fixer_class.order if hasattr(fixer_class, "order") else 0


def select_fixer_classes(
    only_fixers: List[str] = None, except_fixers: List[str] = None
):
    if only_fixers:
        if except_fixers:
            raise ValueError("Cannot use only_fixers and except_fixers together")
        unsorted_iterable = filter(
            lambda fixer_class: _fixer_class_starts_with_prefix(
                fixer_class, only_fixers
            ),
            all_fixer_classes(),
        )
    elif except_fixers:
        unsorted_iterable = filter(
            lambda fixer_class: not _fixer_class_starts_with_prefix(
                fixer_class, except_fixers
            ),
            all_fixer_classes(),
        )
    else:
        unsorted_iterable = all_fixer_classes()
    result = list(unsorted_iterable)
    result.sort(key=try_get_fixer_order)
    return result


def _fixer_class_starts_with_prefix(fixer_class, prefixes: List[str]):
    """
    Checks if the friendly name of the given fixer class starts with any of the given prefixes.
    """
    fixer_key = fixer_class_to_fixer_key(fixer_class)
    for prefix in prefixes:
        if fixer_key.startswith(prefix):
            return True
    return False


def all_fixer_classes():
    """Gets all fixer classes"""
    my_modules = [
        instarepo.fixers.changelog,
        instarepo.fixers.ci,
        instarepo.fixers.dotnet,
        instarepo.fixers.license,
        instarepo.fixers.maven,
        instarepo.fixers.missing_files,
        instarepo.fixers.pascal,
        instarepo.fixers.readme,
        instarepo.fixers.repo_description,
        instarepo.fixers.vb6,
    ]
    for my_module in my_modules:
        my_classes = classes_in_module(my_module)
        for clz in my_classes:
            if clz.__name__.endswith(FIXER_SUFFIX):
                yield clz


def classes_in_module(module):
    """
    Gets the classes defined in the given module
    """
    module_dict = module.__dict__
    return (
        module_dict[c]
        for c in module_dict
        if (
            isinstance(module_dict[c], type)
            and module_dict[c].__module__ == module.__name__
        )
    )



"""Unit tests for discovery.py"""
import collections
import pytest

import instarepo.fixers.base
import instarepo.fixers.changelog
import instarepo.fixers.config
import instarepo.fixers.context
import instarepo.fixers.dotnet
import instarepo.fixers.maven
import instarepo.git

from .discovery import (
    all_fixer_classes,
    classes_in_module,
    select_fixer_classes,
    try_get_fixer_order,
)


@pytest.fixture(params=[clz for clz in all_fixer_classes()])
def fixer_class(request):
    return request.param


def test_fixer_has_doc_string(fixer_class):  # pylint: disable=redefined-outer-name
    """Tests all fixer classes have a doc string"""
    assert fixer_class
    assert fixer_class.__doc__


def test_can_create_fixer(fixer_class):  # pylint: disable=redefined-outer-name
    """Tests that it is possible to instantiate all fixers"""
    mock_git = instarepo.git.GitWorkingDir("/tmp")
    mock_config = instarepo.fixers.config.PartialConfig({})
    mock_github = ()
    repo_type = collections.namedtuple("Repo", ["full_name"])
    mock_repo = repo_type("ngeor/test")
    instance = fixer_class(
        instarepo.fixers.context.Context(
            git=mock_git,
            config=mock_config,
            repo=mock_repo,
            github=mock_github,
            verbose=False,
        )
    )
    assert instance


def test_can_create_fixer_for_local_dir(
    fixer_class,
):  # pylint: disable=redefined-outer-name
    """Tests that it is possible to instantiate all fixers without repo/github instances"""
    mock_git = instarepo.git.GitWorkingDir("/tmp")
    mock_config = instarepo.fixers.config.PartialConfig({})
    instance = fixer_class(
        instarepo.fixers.context.Context(
            git=mock_git, config=mock_config, repo=None, github=None, verbose=False
        )
    )
    assert instance


class TestSelectFixerClasses:
    def test_returns_all_when_unfiltered(self):
        assert len(list(all_fixer_classes())) == len(select_fixer_classes())

    def test_filter_by_name(self):
        assert [
            instarepo.fixers.dotnet.MustHaveCIFix,
        ] == select_fixer_classes(only_fixers=["dotnet"])

    def test_filter_by_two_names(self):
        assert [
            instarepo.fixers.dotnet.MustHaveCIFix,
            instarepo.fixers.maven.MustHaveCIFix,
            instarepo.fixers.maven.MavenBadgesFix,
            instarepo.fixers.maven.UrlFix,
        ] == select_fixer_classes(only_fixers=["dotnet", "maven"])

    def test_cannot_use_only_and_except_together(self):
        with pytest.raises(ValueError):
            select_fixer_classes(only_fixers=["a"], except_fixers=["b"])

    def test_filter_except(self):
        assert [instarepo.fixers.dotnet.MustHaveCIFix,] == select_fixer_classes(
            except_fixers=[
                "changelog",
                "ci",
                "license",
                "maven",
                "missing_files",
                "pascal",
                "r",
                "vb",
            ]
        )

    def test_sort(self):
        assert [
            instarepo.fixers.changelog.MustHaveCliffTomlFix,
            instarepo.fixers.dotnet.MustHaveCIFix,
        ] == select_fixer_classes(only_fixers=["dotnet", "changelog"])


def test_classes_in_module():
    assert [instarepo.git.GitWorkingDir] == list(classes_in_module(instarepo.git))


def test_try_get_fix_order():
    assert try_get_fixer_order(instarepo.fixers.dotnet.MustHaveCIFix) == 0
    assert try_get_fixer_order(instarepo.fixers.changelog.MustHaveCliffTomlFix) == -100



"""Unit tests for config.py"""

from .config import PartialConfig


def test_empty_config():
    config = PartialConfig({})
    assert config.get_setting("ngeor/instarepo", "enabled") is None


def test_missing_repo():
    config = PartialConfig({"defaults": {"enabled": True}})
    assert config.get_setting("ngeor/instarepo", "enabled")


def test_present_repo():
    config = PartialConfig(
        {
            "defaults": {"enabled": True},
            "repos": {"ngeor/instarepo": {"enabled": False}},
        }
    )
    assert not config.get_setting("ngeor/instarepo", "enabled")


def test_nested_key():
    config = PartialConfig(
        {
            "defaults": {"editor": {"overwrite": {True}}},
            "repos": {"ngeor/instarepo": {"enabled": False}},
        }
    )
    assert config.get_setting("ngeor/instarepo", "editor", "overwrite")



from .license import update_copyright_year


def test_no_copyright():
    contents = """
    Does not contain a copyright year
    """
    expected = contents
    actual = update_copyright_year(contents, 2021)
    assert expected == actual


def test_same_year_copyright():
    contents = """
Copyright (c) 2021 Nikolaos Georgiou
    """
    expected = contents
    actual = update_copyright_year(contents, 2021)
    assert expected == actual


def test_same_year_range_copyright():
    contents = """
Copyright (c) 2020-2021 Nikolaos Georgiou
    """
    expected = contents
    actual = update_copyright_year(contents, 2021)
    assert expected == actual


def test_old_year_single_year_copyright():
    contents = """
Copyright (c) 2020 Nikolaos Georgiou
    """
    expected = """
Copyright (c) 2020-2021 Nikolaos Georgiou
    """
    actual = update_copyright_year(contents, 2021)
    assert expected == actual


def test_old_year_year_range_copyright():
    contents = """
Copyright (c) 2019-2020 Nikolaos Georgiou
    """
    expected = """
Copyright (c) 2019-2021 Nikolaos Georgiou
    """
    actual = update_copyright_year(contents, 2021)
    assert expected == actual



"""Unit tests for the context module"""

import collections
from pytest_mock import MockerFixture
from .config import PartialConfig
from .context import Context
from ..git import GitWorkingDir


def test_full_name_from_repo():
    git = ()
    config = PartialConfig({})
    repo_type = collections.namedtuple("Repo", "full_name")
    repo = repo_type("ngeor/test")
    context = Context(git, config, repo, github=None, verbose=False)
    assert context.full_name() == "ngeor/test"


def test_full_name_from_git_remote(mocker: MockerFixture):
    git = GitWorkingDir("/tmp")
    mock = mocker.patch.object(git, "get_remote_url")
    mock.return_value = "git@github.com:foo/bar.git"
    config = PartialConfig({})
    context = Context(git, config, repo=None, github=None, verbose=False)
    assert context.full_name() == "foo/bar"


def test_default_branch_from_repo():
    repo_type = collections.namedtuple("Repo", "default_branch")
    repo = repo_type("main")
    context = Context(None, None, repo=repo)
    assert context.default_branch() == "main"


def test_default_branch_from_git(mocker: MockerFixture):
    git = GitWorkingDir("/tmp")
    mock = mocker.patch.object(git, "get_default_branch")
    mock.return_value = "trunk"
    context = Context(git, None)
    assert context.default_branch() == "trunk"



import requests
from instarepo.fixers.base import MissingFileFix
from .finders import is_lazarus_project, is_maven_project, is_vb6_project
from .naming import fixer_class_to_fixer_key


class MustHaveReadmeFix(MissingFileFix):
    """
    Ensures that the repo has a readme file.

    Does not run for locally checked out repositories.
    """

    def get_filename(self):
        return "README.md"

    def get_contents(self):
        repo = self.context.repo
        contents = f"# {repo.name}\n"
        if repo.description:
            contents = contents + "\n" + repo.description + "\n"
        return contents

    def should_process_repo(self) -> bool:
        repo = self.context.repo
        return repo is not None


EDITOR_CONFIG = """# Editor configuration, see https://editorconfig.org
root = true

[*]
charset = utf-8
indent_style = space
indent_size = 4
insert_final_newline = true
trim_trailing_whitespace = true
max_line_length = 120

[*.sh]
end_of_line = lf

[*.{xml,yml}]
indent_size = 2
"""


class MustHaveEditorConfigFix(MissingFileFix):
    """Ensures an editorconfig file exists"""

    def get_filename(self):
        return ".editorconfig"

    def get_contents(self):
        return EDITOR_CONFIG

    def should_overwrite(self):
        return self.context.get_setting(
            fixer_class_to_fixer_key(self.__class__), "overwrite"
        )


class MustHaveGitHubFundingFix(MissingFileFix):
    """
    Ensures a GitHub funding file exists (.github/FUNDING.yml).
    The template file needs to be configured in the configuration file.

    Does not run for locally checked out repositories.
    """

    def get_filename(self):
        return ".github/FUNDING.yml"

    def should_process_repo(self):
        return (
            self.context.repo
            and not self.context.repo.private
            and not self.context.repo.fork
            and self._get_template_filename()
        )

    def get_contents(self):
        with open(self._get_template_filename(), "r", encoding="utf-8") as file:
            return file.read()

    def _get_template_filename(self):
        return self.context.get_setting("funding.yml")


class MustHaveGitIgnoreFix(MissingFileFix):
    """Ensures a .gitignore file exists"""

    LAZARUS_GITIGNORE = """*.o
*.ppu
*.obj
*.exe
*.dll
*.compiled
*.bak
*.lps
backup/
"""
    VB6_GITIGNORE = """*.exe
*.dll
*.ocx
*.vbw
"""

    def get_filename(self):
        return ".gitignore"

    def get_contents(self):
        if is_maven_project(self.context.git.dir):
            # https://github.com/github/gitignore/blob/master/Maven.gitignore
            response = requests.get(
                "https://raw.githubusercontent.com/github/gitignore/master/Maven.gitignore"
            )
            response.raise_for_status()
            return response.text
        if is_lazarus_project(self.context.git.dir):
            return self.LAZARUS_GITIGNORE
        if is_vb6_project(self.context.git.dir):
            return self.VB6_GITIGNORE
        return None

    def should_overwrite(self):
        return self.context.get_setting(
            fixer_class_to_fixer_key(self.__class__), "overwrite"
        )



"""Fixers for .NET projects"""
import logging
import os
import os.path
from string import Template
from typing import Iterable, List

import instarepo.fixers.context
import instarepo.git
import instarepo.github
import instarepo.xml_utils
from .base import ensure_directories
from .finders import is_file_of_extension
from .naming import fixer_class_to_fixer_key
from ..parsers import (
    many,
    combine_or,
    word,
    one_char_if,
    quoted_string,
    combine_and_opt,
    surrounded_by_space,
    is_symbol,
    until_eol_or_eof,
    is_cr_lf,
    any_char,
)


class MustHaveCIFix:
    """
    Creates a GitHub Action workflow for CSharp projects, deletes appveyor.yml if present.
    """

    def __init__(self, context: instarepo.fixers.context.Context):
        self.context = context

    def run(self):
        sln_paths = list(self._get_sln_paths())
        # multiple sln files not supported
        if len(sln_paths) != 1:
            return []
        sln_path = sln_paths[0]
        cs_projects = [
            self.context.git.join(relative_cs_proj)
            for relative_cs_proj in get_projects_from_sln_file(sln_path)
        ]
        if not cs_projects:
            return []
        rel_sln_path = os.path.relpath(sln_path, self.context.git.dir)
        sln_name, _ = os.path.splitext(os.path.basename(sln_path))
        needs_windows = any(
            filter(lambda x: x == "windows", map(csproj_file_to_os, cs_projects))
        )
        if needs_windows:
            artifact_path = f"{sln_name}*/bin/Release/{sln_name}*.*"
            expected_contents = get_windows_workflow_contents(
                self.context.default_branch(), rel_sln_path, artifact_path
            )
        else:
            artifact_path = f"{sln_name}*/bin/Release/*/{sln_name}*.*"
            expected_contents = get_linux_workflow_contents(
                self.context.default_branch(), artifact_path
            )
        dir_name = ".github/workflows"
        ensure_directories(self.context.git, dir_name)
        file_name = dir_name + "/build.yml"
        absolute_file_name = self.context.git.join(file_name)
        if os.path.isfile(absolute_file_name):
            with open(absolute_file_name, "r", encoding="utf-8") as file:
                old_contents = file.read()
        else:
            old_contents = ""
        if not old_contents or (
            (expected_contents != old_contents) and self._can_overwrite()
        ):
            with open(absolute_file_name, "w", encoding="utf-8") as file:
                file.write(expected_contents)
            self.context.git.add(file_name)
            if old_contents:
                msg = "chore: Updated GitHub Actions workflow for .NET project"
            else:
                msg = "chore: Added GitHub Actions workflow for .NET project"
            self._rm_appveyor()
            self.context.git.commit(msg)
            return [msg]
        if self._rm_appveyor():
            msg = "chore: Removed appveyor.yml from .NET project"
            self.context.git.commit(msg)
            return [msg]
        return []

    def _get_sln_paths(self):
        with os.scandir(self.context.git.dir) as iterator:
            for entry in iterator:
                if is_file_of_extension(entry, ".sln"):
                    yield entry.path

    def _rm_appveyor(self):
        if self.context.git.isfile("appveyor.yml"):
            self.context.git.rm("appveyor.yml")
            return True
        return False

    def _can_overwrite(self):
        return self.context.get_setting(
            fixer_class_to_fixer_key(self.__class__), "overwrite"
        )


def get_linux_workflow_contents(default_branch: str, artifact_path: str):
    template = Template(
        """name: CI

on:
  push:
    branches: [ ${default_branch} ]
  pull_request:
    branches: [ ${default_branch} ]

jobs:
  build:
    runs-on: ubuntu-latest
    steps:
    - uses: actions/checkout@v2
    - name: Set up .NET
      uses: actions/setup-dotnet@v1
      with:
        dotnet-version: '3.1.x'
    - run: dotnet build
    - run: dotnet test -v normal
    - run: dotnet build -c Release
    - name: Upload binaries
      uses: actions/upload-artifact@v3
      with:
        name: binaries
        path: ${artifact_path}
"""
    )
    return template.substitute(
        default_branch=default_branch, artifact_path=artifact_path
    )


def get_windows_workflow_contents(
    default_branch: str, sln_path: str, artifact_path: str
):
    template = Template(
        """name: CI

on:
  push:
    branches: [ ${default_branch} ]
  pull_request:
    branches: [ ${default_branch} ]

jobs:
  build:
    runs-on: windows-latest
    steps:
    - name: Checkout code
      uses: actions/checkout@v2
    - name: Add msbuild to PATH # https://github.com/microsoft/setup-msbuild
      uses: microsoft/setup-msbuild@v1.1
    - name: Setup NuGet # https://github.com/NuGet/setup-nuget
      uses: nuget/setup-nuget@v1
    - name: Restore NuGet packages
      run: nuget restore ${sln_path}
    - name: Build project
      run: msbuild -t:rebuild -property:Configuration=Release ${sln_path}
    - name: Upload binaries
      uses: actions/upload-artifact@v3
      with:
        name: binaries
        path: ${artifact_path}
"""
    )
    return template.substitute(
        default_branch=default_branch, sln_path=sln_path, artifact_path=artifact_path
    )


def get_projects_from_sln_file(path: str) -> List[str]:
    """
    Gets the projects defined in a sln file.

    :param path: The path of a Visual Studio sln file.
    """
    with open(path, "r", encoding="utf-8") as file:
        return list(get_projects_from_sln_file_contents(file.read()))


def get_projects_from_sln_file_contents(contents: str) -> Iterable[str]:
    """
    Gets the projects defined in a sln file.

    :param contents: The contents of a Visual Studio sln file.
    """
    return SlnProjectFinder(contents)


class SlnProjectFinder:
    def __init__(self, contents: str):
        self._parser = SlnParser(contents)

    def next(self):
        while self._parser.find("Project"):
            project_path = self._read_project_path()
            if project_path:
                return project_path

    def _read_project_path(self):
        lparen = self._parser.next()
        if lparen != "(":
            return
        project_type_guid = self._parser.next()
        if project_type_guid[1:-1] not in [
            "{FAE04EC0-301F-11D3-BF4B-00C04F79EFBC}",
            "{9A19103F-16F7-4668-BE54-9A1E7A4F7556}",
        ]:
            return
        rparen = self._parser.next()
        if rparen != ")":
            return
        eq_sign = self._parser.next()
        if eq_sign != "=":
            return
        _project_name = self._parser.next()
        comma = self._parser.next()
        if comma != ",":
            return
        csproj_path = self._parser.next()
        return csproj_path[1:-1]

    def __iter__(self):
        return self

    def __next__(self):
        result = self.next()
        if result:
            return result
        else:
            raise StopIteration


class SlnParser:
    def __init__(self, contents: str):
        self._contents = contents
        self._parser = combine_or(
            comment(),
            word(),
            version_number(),
            quoted_string(),
            surrounded_by_space(one_char_if(is_symbol)),
            many(one_char_if(is_cr_lf)),
            any_char(),
        )

    def next(self):
        result, remaining = self._parser(self._contents)
        self._contents = remaining
        return result

    def find(self, needle: str):
        """
        Returns the first token that is equal to the parameter.
        """
        token = self.next()
        while token and token != needle:
            token = self.next()
        return token


def version_number():
    return many(one_char_if(lambda char: char == "." or (char >= "0" and char <= "9")))


def comment():
    return combine_and_opt(one_char_if(lambda ch: ch == "#"), until_eol_or_eof())


def csproj_file_to_os(csproj_filename):
    tree = instarepo.xml_utils.parse(csproj_filename)
    if tree is None:
        logging.warning("Could not parse %s", csproj_filename)
        return
    return csproj_root_node_to_os(tree.getroot(), csproj_filename)


def csproj_root_node_to_os(root_node, csproj_filename=""):
    """
    Determines the OS needed by the given csproj file. The root_node
    is the root node of the parsed XML file.
    """
    if root_node is None:
        logging.warning("No root node for csproj file %s", csproj_filename)
        return
    if root_node.tag == "{http://schemas.microsoft.com/developer/msbuild/2003}Project":
        return "windows"
    node = instarepo.xml_utils.find(root_node, "PropertyGroup", "TargetFramework")
    if node is None:
        logging.warning("Could not find target framework of %s", csproj_filename)
        return
    if not node.text:
        logging.warning("Empty TargetFramework for %s", csproj_filename)
        return
    if node.text.startswith("net4"):
        return "windows"
    if node.text.startswith("netcore") or node.text.startswith("netstandard"):
        return "linux"



"""Unit tests for missing_files.py"""

import os.path
import tempfile
from pytest_mock import MockerFixture
import instarepo.fixers.config
import instarepo.fixers.context
from .missing_files import MustHaveEditorConfigFix, EDITOR_CONFIG
from ..git import GitWorkingDir


def test_editor_config(mocker: MockerFixture):
    with tempfile.TemporaryDirectory() as tmp_dir:
        # arrange
        filename = os.path.join(tmp_dir, ".editorconfig")
        git = GitWorkingDir(tmp_dir, quiet=True)
        git_add = mocker.patch.object(git, "add")
        git_commit = mocker.patch.object(git, "commit")
        config = instarepo.fixers.config.PartialConfig({})
        context = instarepo.fixers.context.Context(git, config)
        fixer = MustHaveEditorConfigFix(context)

        # act
        result = fixer.run()

        # assert
        assert result == ["chore: Adding .editorconfig"]
        assert os.path.isfile(os.path.join(tmp_dir, ".editorconfig"))
        with open(filename, "r", encoding="utf-8") as file:
            assert file.read() == EDITOR_CONFIG
        git_add.assert_called_once_with(".editorconfig")
        git_commit.assert_called_once_with("chore: Adding .editorconfig")


def test_editor_config_already_exists(mocker: MockerFixture):
    with tempfile.TemporaryDirectory() as tmp_dir:
        # arrange
        filename = os.path.join(tmp_dir, ".editorconfig")
        with open(filename, "w", encoding="utf-8") as file:
            file.write("# hello world")
        git = GitWorkingDir(tmp_dir, quiet=True)
        git_add = mocker.patch.object(git, "add")
        git_commit = mocker.patch.object(git, "commit")
        git_get_remote_url = mocker.patch.object(git, "get_remote_url")
        git_get_remote_url.return_value = "git@github.com:ngeor/instarepo.git"
        config = instarepo.fixers.config.PartialConfig({})
        context = instarepo.fixers.context.Context(git, config)
        fixer = MustHaveEditorConfigFix(context)

        # act
        result = fixer.run()

        # assert
        assert result == []
        assert os.path.isfile(os.path.join(tmp_dir, ".editorconfig"))
        with open(filename, "r", encoding="utf-8") as file:
            assert file.read() == "# hello world"
        git_add.assert_not_called()
        git_commit.assert_not_called()


def test_editor_config_already_exists_overwrite(mocker: MockerFixture):
    with tempfile.TemporaryDirectory() as tmp_dir:
        # arrange
        filename = os.path.join(tmp_dir, ".editorconfig")
        with open(filename, "w", encoding="utf-8") as file:
            file.write("# hello world")
        git = GitWorkingDir(tmp_dir, quiet=True)
        git_add = mocker.patch.object(git, "add")
        git_commit = mocker.patch.object(git, "commit")
        git_get_remote_url = mocker.patch.object(git, "get_remote_url")
        git_get_remote_url.return_value = "git@github.com:ngeor/instarepo.git"
        config = instarepo.fixers.config.PartialConfig(
            {"defaults": {"missing_files.must_have_editor_config": {"overwrite": True}}}
        )
        context = instarepo.fixers.context.Context(git, config)
        fixer = MustHaveEditorConfigFix(context)

        # act
        result = fixer.run()

        # assert
        assert result == ["chore: Updated .editorconfig"]
        assert os.path.isfile(os.path.join(tmp_dir, ".editorconfig"))
        with open(filename, "r", encoding="utf-8") as file:
            assert file.read() == EDITOR_CONFIG
        git_add.assert_called_once_with(".editorconfig")
        git_commit.assert_called_once_with("chore: Updated .editorconfig")


def test_editor_config_already_exists_matching_template(mocker: MockerFixture):
    with tempfile.TemporaryDirectory() as tmp_dir:
        # arrange
        filename = os.path.join(tmp_dir, ".editorconfig")
        with open(filename, "w", encoding="utf-8") as file:
            file.write(EDITOR_CONFIG)
        git = GitWorkingDir(tmp_dir, quiet=True)
        git_add = mocker.patch.object(git, "add")
        git_commit = mocker.patch.object(git, "commit")
        git_get_remote_url = mocker.patch.object(git, "get_remote_url")
        git_get_remote_url.return_value = "git@github.com:ngeor/instarepo.git"
        config = instarepo.fixers.config.PartialConfig(
            {"defaults": {"missing_files.must_have_editor_config": {"overwrite": True}}}
        )
        context = instarepo.fixers.context.Context(git, config)
        fixer = MustHaveEditorConfigFix(context)

        # act
        result = fixer.run()

        # assert
        assert result == []
        assert os.path.isfile(os.path.join(tmp_dir, ".editorconfig"))
        with open(filename, "r", encoding="utf-8") as file:
            assert file.read() == EDITOR_CONFIG
        git_add.assert_not_called()
        git_commit.assert_not_called()



"""
Fixers regarding CI.
"""

import os.path
import re
from string import Template

import instarepo.fixers.base
import instarepo.fixers.context


class NoTravisFix:
    """Removes the .travis.yml file"""

    def __init__(self, context: instarepo.fixers.context.Context):
        self.context = context

    def run(self):
        filename = ".travis.yml"
        full_name = self.context.git.join(filename)
        if not os.path.isfile(full_name):
            return []
        self.context.git.rm(filename)
        msg = f"chore: Removed {filename}"
        self.context.git.commit(msg)
        return [msg]


class NoTravisBadgeFix(instarepo.fixers.base.SingleFileFix):
    """Removes the Travis badge from README files"""

    def __init__(self, context: instarepo.fixers.context.Context):
        super().__init__(
            context.git, "README.md", "chore: Removed Travis badge from README"
        )

    def convert(self, contents: str) -> str:
        return remove_travis_badge(contents)


RE_BADGE = re.compile(
    r"\[!\[Build Status\]\(https://travis-ci[^)]+\)\]\(https://travis-ci[^)]+\)"
)


def remove_travis_badge(contents: str) -> str:
    return RE_BADGE.sub("", contents)


class PythonBuildFix(instarepo.fixers.base.MissingFileFix):
    """Adds a build GitHub action for Python projects"""

    def get_filename(self):
        return ".github/workflows/build.yml"

    def get_contents(self) -> str:
        with open(self.context.git.join("Pipfile"), "r", encoding="utf-8") as file:
            pipfile_lines = file.readlines()
        has_build_dependency = any(
            filter(lambda line: line.startswith("build"), pipfile_lines)
        )
        if has_build_dependency:
            return """name: Python CI

on:
  push:
    branches: [ trunk ]
  pull_request:
    branches: [ trunk ]

jobs:
  build:
    runs-on: ubuntu-latest
    strategy:
      matrix:
        python-version: ['3.8', '3.9', '3.10']
    steps:
      - uses: actions/checkout@v2
      - name: Set up Python ${{ matrix.python-version }}
        uses: actions/setup-python@v2
        with:
          python-version: ${{ matrix.python-version }}
      - name: Install dependencies
        run: |
          python -m pip install --upgrade pip
          pip install pipenv
          pipenv install --dev
      - name: Test with pytest
        run: |
          pipenv run pytest
      - name: Build wheel
        run: |
          pipenv run python -m build
      - name: Upload wheel
        uses: actions/upload-artifact@v2
        with:
          name: wheel-${{ matrix.python-version }}
          path: dist/*.whl
"""

        return """name: Python CI

on:
  push:
    branches: [ trunk ]
  pull_request:
    branches: [ trunk ]
jobs:
  build:
    runs-on: ubuntu-latest
    strategy:
      matrix:
        python-version: ['3.8', '3.9', '3.10']
    steps:
      - uses: actions/checkout@v2
      - name: Set up Python ${{ matrix.python-version }}
        uses: actions/setup-python@v2
        with:
          python-version: ${{ matrix.python-version }}
      - name: Install dependencies
        run: |
          python -m pip install --upgrade pip
          pip install pipenv
          pipenv install --dev
      - name: Test with pytest
        run: |
          pipenv run pytest
"""

    def should_process_repo(self) -> bool:
        return self.context.git.isfile("Pipfile")


class PythonReleaseFix(instarepo.fixers.base.MissingFileFix):
    """Adds a release GitHub action for Python projects"""

    def get_filename(self):
        return ".github/workflows/release.yml"

    def get_contents(self) -> str:
        template = Template(
            """name: Release

on:
  push:
    tags: [ "v*" ]

jobs:
  release:
    runs-on: ubuntu-latest
    steps:
      - uses: actions/checkout@v2
      - name: Set up Python
        uses: actions/setup-python@v2
        with:
          python-version: '3.8'
      - name: Install dependencies
        run: |
          python -m pip install --upgrade pip
          pip install pipenv
          pipenv install --dev
      - name: Test with pytest
        run: |
          pipenv run pytest
      - name: Build wheel
        run: |
          pipenv run python -m build
      - name: Publish wheel with twine
        run: >
          pipenv run twine upload
          -u $${{secrets.${twine_username_variable}}}
          -p $${{secrets.${twine_password_variable}}}
          --non-interactive
          --disable-progress-bar
          dist/*
"""
        )
        name = self.context.full_name().split("/")[-1].upper()
        return template.substitute(
            twine_username_variable=f"{name}_TWINE_USERNAME",
            twine_password_variable=f"{name}_TWINE_PASSWORD",
        )

    def should_process_repo(self) -> bool:
        if not self.context.git.isfile("Pipfile"):
            return False
        with open(self.context.git.join("Pipfile"), "r", encoding="utf-8") as file:
            pipfile_lines = file.readlines()
        has_twine_dependency = any(
            filter(lambda line: line.startswith("twine"), pipfile_lines)
        )
        return has_twine_dependency



"""
Fixers for the README.md file
"""

import os.path
import re
from typing import List, Tuple

import instarepo.fixers.context
import instarepo.git
from .base import SingleFileFix

# \w matches [a-zA-Z0-9_]
RE_MARKDOWN_IMAGE = re.compile(
    r'!\[[^]]*\]\((?P<filename>/[\w/\.]+)\?raw=true "[^"]*"\)'
)


class ReadmeImageFix(SingleFileFix):
    """
    Finds broken images in the `README.md` file.
    Able to correct images that were moved one or more
    folders up but the user forgot to update them in the `README.md` file.
    """

    def __init__(self, context: instarepo.fixers.context.Context):
        """Creates an instance of this class"""
        super().__init__(context.git, "README.md", "fix: Fixed broken images in README")

    def convert(self, contents: str) -> str:
        return RE_MARKDOWN_IMAGE.sub(self.image_convert, contents)

    def image_convert(self, match: re.Match) -> str:
        filename = match.group("filename")
        new_filename = self.find_new_filename(filename)
        return (
            match.string[match.start() : match.start("filename")]
            + new_filename
            + match.string[match.end("filename") : match.end()]
        )

    def find_new_filename(self, filename: str) -> str:
        """
        Finds a new filename for the broken image.
        Removes leading parts of the given path until a matching
        filename is found.
        """
        abs_git_dir = os.path.abspath(self.git.dir)
        parts = [x for x in filename.replace("\\", "/").split("/") if x]
        while parts:
            possible_path = os.path.join(abs_git_dir, *parts)
            if os.path.isfile(possible_path):
                return "/" + "/".join(parts)
            parts = parts[1:]
        return filename


def locate_badges(readme_contents: str) -> Tuple[str, List[str], str]:
    before_badges = []
    badges = []
    after_badges = []
    state = 0
    for line in readme_contents.splitlines():
        if state == 0:
            if _is_badge_line(line):
                state = 1
                badges.append(line)
            else:
                before_badges.append(line)
        elif state == 1:
            if _is_badge_line(line):
                badges.append(line)
            else:
                state = 2
                after_badges.append(line)
        else:
            after_badges.append(line)
    if state == 0:
        # no badges were found, try to split based on headline
        return (_join_lines(before_badges[0:1]), [], _join_lines(before_badges[1:]))
    return (_join_lines(before_badges), badges, _join_lines(after_badges))


def _is_badge_line(line: str) -> bool:
    return line.startswith("[![")


def _join_lines(lines: List[str]) -> str:
    return "\n".join(_strip_empty_lines(lines)) + "\n"


def _strip_empty_lines(lines: List[str]) -> List[str]:
    return _strip_trailing_empty_lines(_strip_leading_empty_lines(lines))


def _strip_leading_empty_lines(lines: List[str]) -> List[str]:
    i = 0
    while i < len(lines) and len(lines[i]) == 0:
        i += 1
    return lines[i:]


def _strip_trailing_empty_lines(lines: List[str]) -> List[str]:
    i = len(lines) - 1
    while i >= 0 and len(lines[i]) == 0:
        i -= 1
    return lines[0 : i + 1]


def merge_badges(before_text: str, badges: List[str], after_text: str) -> str:
    if badges:
        return before_text + "\n" + "\n".join(badges) + "\n\n" + after_text
    return before_text + "\n" + after_text



from instarepo.fixers.base import MissingFileFix


class MustHaveCliffTomlFix(MissingFileFix):
    """Ensures the configuration for git-cliff (cliff.toml) exists"""

    order = -100

    def should_process_repo(self):
        return self._get_template_filename()

    def get_contents(self):
        with open(self._get_template_filename(), "r", encoding="utf-8") as file:
            return file.read()

    def get_filename(self):
        return "cliff.toml"

    def _get_template_filename(self):
        return self.context.get_setting("cliff.toml")



FIXER_PREFIX = "instarepo.fixers."
FIXER_SUFFIX = "Fix"


def fixer_class_to_fixer_key(clz):
    """
    Derives the unique fixer identifier out of a fixer class.
    The identifier is shorter and can be used to dynamically
    turn fixers on/off via the CLI.
    """
    full_module_name: str = clz.__module__
    expected_prefix = FIXER_PREFIX
    if not full_module_name.startswith(expected_prefix):
        raise ValueError(
            f"Module {full_module_name} did not start with prefix {expected_prefix}"
        )
    expected_suffix = FIXER_SUFFIX
    if not clz.__name__.endswith(expected_suffix):
        raise ValueError(
            f"Module {clz.__name__} did not end with suffix {expected_suffix}"
        )
    my_module = full_module_name[len(expected_prefix) :]
    return (
        my_module
        + "."
        + _pascal_case_to_underscore_case(clz.__name__[0 : -len(expected_suffix)])
    )


def _pascal_case_to_underscore_case(value: str) -> str:
    """
    Converts a pascal case string (e.g. MyClass)
    into a lower case underscore separated string (e.g. my_class).
    """
    result = ""
    state = "initial"
    partial = ""
    for char in value:
        if "A" <= char <= "Z":
            if state == "initial":
                state = "upper"
            elif state == "upper":
                state = "multi-upper"
            else:
                if result:
                    result += "_"
                result += partial
                partial = ""
                state = "upper"
            partial += char.lower()
        else:
            if state == "multi-upper":
                if result:
                    result += "_"
                result += partial
                partial = ""
            partial += char
            state = "lower"

    if result:
        result += "_"
    result += partial
    return result



from .parsers import any_char, one_char_if, many, word, combine_and


def test_any_char():
    parsed, remaining = any_char()("abc")
    assert parsed == "a"
    assert remaining == "bc"


def test_one_char_if():
    parsed, remaining = one_char_if(lambda p: p >= "a")("abc")
    assert parsed == "a"
    assert remaining == "bc"
    parsed, remaining = one_char_if(lambda p: p < "a")("abc")
    assert parsed == ""
    assert remaining == "abc"


def test_many():
    lower_case_parser = one_char_if(lambda x: x >= "a" and x <= "z")
    parsed, remaining = many(lower_case_parser)("hello world")
    assert parsed == "hello"
    assert remaining == " world"


def test_word():
    parsed, remaining = word()("Hello, world!")
    assert parsed == "Hello"
    assert remaining == ", world!"


def test_combine_and():
    first = word()
    second = one_char_if(lambda x: x == ",")
    parsed, remaining = combine_and(first, second)("Hello, world!")
    assert parsed == "Hello,"
    assert remaining == " world!"



"""Unit tests for naming.py"""

import instarepo.fixers.dotnet
import instarepo.fixers.maven

from .naming import fixer_class_to_fixer_key


def test_fixer_class_to_fixer_key():
    """Tests various fixer classes can be mapped to a key"""
    assert (
        fixer_class_to_fixer_key(instarepo.fixers.dotnet.MustHaveCIFix)
        == "dotnet.must_have_ci"
    )
    assert fixer_class_to_fixer_key(instarepo.fixers.maven.UrlFix) == "maven.url"



import json
import os.path


def load_config(filename: str):
    default_config = load_default_config()
    if filename:
        user_config = load_partial_config(filename)
        return Config(user_config, default_config)
    else:
        return default_config


class PartialConfig:
    def __init__(self, config):
        self._config = config

    def get_setting(self, full_name: str, *args):
        repo_value = _lookup_key(self._config, "repos", full_name, *args)
        if repo_value is None:
            return _lookup_key(self._config, "defaults", *args)
        else:
            return repo_value


class Config:
    def __init__(self, user_config: PartialConfig, default_config: PartialConfig):
        self._user_config = user_config
        self._default_config = default_config

    def get_setting(self, full_name: str, *args):
        result = self._user_config.get_setting(full_name, *args)
        if result is None:
            return self._default_config.get_setting(full_name, *args)
        else:
            return result


def load_default_config():
    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, "config.json")
    return load_partial_config(filename)


def load_partial_config(filename):
    with open(filename, "r", encoding="utf-8") as file:
        return PartialConfig(json.load(file))


def _lookup_key(dictionary, *args):
    if not dictionary:
        return None
    current = dictionary
    for key in args:
        if not key in current:
            return None
        current = current[key]
    return current



"""
Fetches repository info from GitHub.
"""

from __future__ import annotations
from datetime import datetime, timedelta, timezone
from typing import Iterable, Optional
from enum import Enum, auto, unique

import instarepo.github

from .credentials import build_requests_auth


@unique
class FilterMode(Enum):
    """
    Controls filtering of repositories based on a boolean property (e.g. 'archived').

    ALLOW: Allows repositories, regardless of the property value
    DENY: Includes only repositories where the property is false
    ONLY: Includes only repositories where the property is true
    """

    ALLOW = auto()
    DENY = auto()
    ONLY = auto()

    @staticmethod
    def parse(value: str):
        """
        Parses the given string value and returns the FilterMode.
        Matching is case sensitive. If not found, an error will be thrown.
        """
        return [member for member in FilterMode if member.name == value][0]


class StringFilter:
    """
    Defines a string filter.
    Allows filtering repositories based on a string property.
    The combination of `value` and `mode` allows to
    include or exclude repositories based on the value of
    the property that is being filtered on.
    """

    def __init__(self, value: str = "", mode: FilterMode = FilterMode.ALLOW):
        self.value = value
        self.mode = mode


class RepoSource:
    """
    Retrieves repository information from GitHub.
    """

    def __init__(
        self,
        github: instarepo.github.GitHub,
        sort: str,
        direction: str,
        archived: FilterMode,
        forks: FilterMode,
        repo_prefix: StringFilter,
        language: StringFilter,
        pushed_after,
        pushed_before,
    ):
        """
        Creates an instance of this class

        Parameters:

        :param github: The instance of the GitHub client
        :param sort: The field to sort by
        :param direction: The direction to sort by
        :param archived: Determines how to filter archived repositories
        :param forks: Determines how to filter forks
        :param repo_prefix: Optionally filter repositories whose name starts with this prefix
        :param language: Optionally filter repositories by their language
        :param pushed_after: Optionally filter repositories that were pushed after the given timedelta
        :param pushed_before: Optionally filter repositories that were pushed before the given timedelta
        """
        self.github = github
        self.sort = sort
        self.direction = direction
        self.archived = archived
        self.forks = forks
        self.repo_prefix = repo_prefix
        self.language = language
        self.pushed_after = pushed_after
        self.pushed_before = pushed_before

    def get(self) -> Iterable[instarepo.github.Repo]:
        """
        Retrieves repository information from GitHub.
        """
        return self._filter_pushed_after(
            self._filter_pushed_before(
                self._filter_language(
                    self._filter_prefix(
                        self._filter_forks(
                            self._filter_archived(
                                self.github.get_all_repos(self.sort, self.direction)
                            )
                        )
                    )
                )
            )
        )

    def _filter_archived(self, repos: Iterable[instarepo.github.Repo]):
        if self.archived == FilterMode.ONLY:
            return (repo for repo in repos if repo.archived)
        elif self.archived == FilterMode.DENY:
            return (repo for repo in repos if not repo.archived)
        else:
            return repos

    def _filter_forks(self, repos: Iterable[instarepo.github.Repo]):
        if self.forks == FilterMode.ONLY:
            return (repo for repo in repos if repo.fork)
        elif self.forks == FilterMode.DENY:
            return (repo for repo in repos if not repo.fork)
        else:
            return repos

    def _filter_prefix(self, repos: Iterable[instarepo.github.Repo]):
        return filter_by_name_prefix(repos, self.repo_prefix)

    def _filter_language(self, repos: Iterable[instarepo.github.Repo]):
        return filter_by_language(repos, self.language)

    def _filter_pushed_after(self, repos: Iterable[instarepo.github.Repo]):
        if self.pushed_after:
            return (
                repo
                for repo in repos
                if repo.pushed_at + self.pushed_after > datetime.now(timezone.utc)
            )
        else:
            return repos

    def _filter_pushed_before(self, repos: Iterable[instarepo.github.Repo]):
        if self.pushed_before:
            return (
                repo
                for repo in repos
                if repo.pushed_at + self.pushed_before < datetime.now(timezone.utc)
            )
        else:
            return repos


class RepoSourceBuilder:
    """
    A builder for RepoSource instances.
    """

    def __init__(self):
        """
        Creates an instance of this class.

        """
        self.github = None
        self.sort: str = ""
        self.direction: str = ""
        self.archived = FilterMode.DENY
        self.forks = FilterMode.DENY
        self.repo_prefix = StringFilter()
        self.language = StringFilter()
        self.pushed_after = None
        self.pushed_before = None

    def with_github(self, github: instarepo.github.GitHub):
        """
        Uses the given GitHub client.
        """
        self.github = github
        return self

    def with_args(self, args) -> RepoSourceBuilder:
        """
        Uses the properties defined in the given CLI arguments.

        If the github client is already set with the `with_github`
        method, it is not overwritten. Otherwise, it creates
        a read-only GitHub client.
        """
        if self.github is None:
            self.github = instarepo.github.GitHub(auth=build_requests_auth(args))
        if "sort" in args:
            self.sort = args.sort
        if "direction" in args:
            self.direction = args.direction
        self.forks = FilterMode.parse(args.forks.upper())
        if "archived" in args:
            self.archived = FilterMode.parse(args.archived.upper())

        if args.only_name_prefix:
            self.repo_prefix = StringFilter(args.only_name_prefix, FilterMode.ONLY)
        elif args.except_name_prefix:
            self.repo_prefix = StringFilter(args.except_name_prefix, FilterMode.DENY)

        if args.only_language:
            self.language = StringFilter(args.only_language, FilterMode.ONLY)
        elif args.except_language:
            self.language = StringFilter(args.except_language, FilterMode.DENY)

        self.pushed_after = parse_timedelta(args.pushed_after)
        self.pushed_before = parse_timedelta(args.pushed_before)

        return self

    def build(self):
        """
        Builds a new `RepoSource` instance.
        """
        if self.github is None:
            raise ValueError("GitHub client is mandatory")
        return RepoSource(
            self.github,
            self.sort,
            self.direction,
            self.archived,
            self.forks,
            self.repo_prefix,
            self.language,
            self.pushed_after,
            self.pushed_before,
        )


def filter_by_name_prefix(
    repos: Iterable[instarepo.github.Repo], string_filter: Optional[StringFilter]
) -> Iterable[instarepo.github.Repo]:
    """
    Filters the given repos on their name using the given filter.
    """
    if (
        not string_filter
        or not string_filter.value
        or string_filter.mode == FilterMode.ALLOW
    ):
        return repos
    if string_filter.mode == FilterMode.ONLY:
        return (repo for repo in repos if repo.name.startswith(string_filter.value))
    elif string_filter.mode == FilterMode.DENY:
        return (repo for repo in repos if not repo.name.startswith(string_filter.value))
    else:
        raise ValueError("Invalid filter mode " + string_filter.mode)


def filter_by_language(
    repos: Iterable[instarepo.github.Repo], string_filter: StringFilter
) -> Iterable[instarepo.github.Repo]:
    """
    Filters the given repos on their language using the given filter.
    """
    if not string_filter or string_filter.mode == FilterMode.ALLOW:
        return repos
    if string_filter.mode == FilterMode.ONLY:
        if string_filter.value:
            return (repo for repo in repos if repo.language == string_filter.value)
        else:
            return (repo for repo in repos if not repo.language)
    elif string_filter.mode == FilterMode.DENY:
        if string_filter.value:
            return (repo for repo in repos if repo.language != string_filter.value)
        else:
            return (repo for repo in repos if repo.language)
    else:
        raise ValueError("Invalid filter mode " + string_filter.mode)


def parse_timedelta(value: Optional[str]):
    """
    Parses a time delta string value. For example "15m" is parsed as a
    timedelta of 15 minutes.
    """
    if not value:
        return None
    unit = value[-1]
    amount = int(value[0:-1])
    if unit == "h":
        return timedelta(hours=amount)
    elif unit == "m":
        return timedelta(minutes=amount)
    elif unit == "d":
        return timedelta(days=amount)
    else:
        raise ValueError(f"Invalid time unit: {value}")



"""Unit tests for fix.py"""
import os

import instarepo.fixers.base
import instarepo.fixers.changelog
import instarepo.fixers.config
import instarepo.fixers.context
import instarepo.fixers.dotnet
import instarepo.fixers.maven
import instarepo.git

from .fix import (
    create_composite_fixer,
    format_body,
)


class TestFormatBody:
    """Unit tests for format_body"""

    def test_one_change(self):
        changes = ["Simple change"]
        expected_body = """The following fixes have been applied:
- Simple change
""".replace(
            os.linesep, "\n"
        )
        actual_body = format_body(changes)
        assert actual_body == expected_body

    def test_two_changes(self):
        changes = ["Simple change", "Second change"]
        expected_body = """The following fixes have been applied:
- Simple change
- Second change
""".replace(
            os.linesep, "\n"
        )
        actual_body = format_body(changes)
        assert actual_body == expected_body

    def test_convert_multi_line_to_indentation(self):
        changes = [
            """Complex change
Updated parent to 1.0
"""
        ]
        expected_body = """The following fixes have been applied:
- Complex change
  Updated parent to 1.0
""".replace(
            os.linesep, "\n"
        )
        actual_body = format_body(changes)
        assert actual_body == expected_body


def test_create_composite_fixer():
    # arrange
    context = instarepo.fixers.context.Context(git=None, config=None)
    fixer_classes = [
        instarepo.fixers.changelog.MustHaveCliffTomlFix,
    ]
    # act
    composite_fixer = create_composite_fixer(fixer_classes, context)
    # assert
    assert composite_fixer
    assert isinstance(composite_fixer, instarepo.fixers.base.CompositeFix)
    assert composite_fixer.rules
    assert isinstance(
        composite_fixer.rules[0], instarepo.fixers.changelog.MustHaveCliffTomlFix
    )



def quote():
    return one_char_if(lambda char: char == '"')


def not_quote():
    return many(one_char_if(lambda char: char != '"'))


def quoted_string():
    return combine_and(combine_and(quote(), not_quote()), quote())


def whitespace():
    return many(one_char_if(is_whitespace))


def surrounded_by_space(parser):
    def _surrounded_by_space(contents: str):
        _, remaining = whitespace()(contents)
        result, remaining = parser(remaining)
        if result:
            _, remaining = whitespace()(remaining)
            return result, remaining
        return "", contents

    return _surrounded_by_space


def is_symbol(char):
    return char in ",()="


def is_whitespace(char):
    return char == " " or char == "\t"


def any_char():
    def _any_char(contents: str):
        if len(contents) > 0:
            result = contents[0]
            remaining = contents[1:]
            return result, remaining
        else:
            return "", ""

    return _any_char


def one_char_if(predicate):
    def _one_char_if(contents: str):
        result, remaining = any_char()(contents)
        if predicate(result):
            return result, remaining
        else:
            return "", contents

    return _one_char_if


def many(parser):
    def _many(contents: str):
        eof = False
        remaining = contents
        total_result = ""
        while not eof:
            result, remaining = parser(remaining)
            total_result += result
            eof = not result
        return total_result, remaining

    return _many


def is_letter(char):
    return (char >= "a" and char <= "z") or (char >= "A" and char <= "Z")


def letter():
    return one_char_if(is_letter)


def word():
    return many(letter())


def combine_and(first, second):
    """
    Parser combinator where both first and second parser
    must return a successful result.
    """

    def _combine_and(contents: str):
        first_result, remaining = first(contents)
        if first_result:
            second_result, remaining = second(remaining)
            if second_result:
                return first_result + second_result, remaining
        return "", contents

    return _combine_and


def combine_and_opt(first, second):
    def _combine_and_opt(contents: str):
        first_result, remaining = first(contents)
        if first_result:
            second_result, remaining = second(remaining)
            return first_result + second_result, remaining
        return "", contents

    return _combine_and_opt


def combine_or(*parsers):
    def _combine_or(contents: str):
        for parser in parsers:
            result, remaining = parser(contents)
            if result:
                return result, remaining
        return "", contents

    return _combine_or


def is_cr_lf(char):
    return char == "\r" or char == "\n"


def until_eol_or_eof():
    def _until_eol_or_eof(contents: str):
        i = 0
        while i < len(contents) and not is_cr_lf(contents[i]):
            i = i + 1
        j = i
        while j < len(contents) and is_cr_lf(contents[j]):
            j = j + 1
        return contents[0:i], contents[j:]

    return _until_eol_or_eof
